/*
 * Copyright 2015 Jean-Michel RICHER
 *
 * This program represents an example of CUDA programming.
 *
 * We compute the sum
 */
#include <stdio.h>
#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <iterator>
#include <algorithm>
#include <numeric>
#include <cassert>
using namespace std;
#include <cuda.h>



#include "parameters.h"
#include "cpu_timer.h"
#include "gpu_timer.h"

// ==================================================================
// program command line arguments
// ==================================================================
static struct option long_options[] = {
		{"gpu"  	, required_argument, 0, 'g'},
		{"repeat"	, required_argument, 0, 'r'},
		{"size"		, required_argument, 0, 's'},
		{"verbose"	, no_argument, 0, 'v'},
		{0,0,0,0}
};


/**
 * This macro checks return value of the CUDA runtime call and exits
 * the application if the call failed.
 */
#define cuda_check(value) {     \
		cudaError_t err = value; \
		if (err != cudaSuccess) {       \
			cerr << endl; \
			cerr << "============================================\n"; \
			cerr << "Error: " << cudaGetErrorString(err) << " at line "; \
			cerr << __LINE__ << " in file " <<  __FILE__;   \
			cerr <<  endl; \
			exit(EXIT_FAILURE); \
		} \
}

#define cuda_check_kernel() { \
		cudaError_t err = cudaGetLastError(); \
		if (err != cudaSuccess)  { \
			cerr << endl; \
			cerr << "============================================\n"; \
			cerr << "Kernel Error: " << cudaGetErrorString(err) << " at line "; \
			cerr << __LINE__ << " in file " <<  __FILE__;   \
			cerr <<  endl; \
			exit(EXIT_FAILURE); \
		} \
}

__global__ void kernel_sum(float *x, float *y, float *z, int size) {

	int i = threadIdx.x ;
	if (i < size) {
		z[i] = x[i] + y[i]; // paralell part
	}
}


/**
 * main function
 */
int main(int argc, char *argv[]) {
	// number of GPU on this computer
	int gpu_count = -1;

	Parameters& parameters = Parameters::get_instance();

	// count the number of devices on the machine
	cuda_check(cudaGetDeviceCount(&gpu_count));

	if (gpu_count == 0) {
		cerr << "No CUDA devices found\n";
		return EXIT_FAILURE;
	}


	parameters.setup(argc, argv, long_options, gpu_count);
	cout << parameters << endl;

	cuda_check(cudaSetDevice(parameters.gpu_id));
	cudaDeviceProp properties;
	cuda_check(cudaGetDeviceProperties(&properties, parameters.gpu_id));


	// allocate data on computer global memory (host)
	float *c_x, *c_y, *c_z;

	c_x = new float[ parameters.size ];
	c_y = new float[ parameters.size ];
	c_z = new float[ parameters.size ];

	// allocate data on GPU global memory
	float *g_x, *g_y, *g_z;

	cuda_check(cudaMalloc( (void**) &g_x, parameters.size * sizeof(float) ));
	cuda_check(cudaMalloc( (void**) &g_y, parameters.size * sizeof(float) ));
	cuda_check(cudaMalloc( (void**) &g_z, parameters.size * sizeof(float) ));


	cuda_check(cudaDeviceSynchronize());

	CPUTimer c_timer;
	GPUTimer g_timer;

	dim3 grid_dim(1, 1, 1);
	dim3 block_dim(parameters.size, 1, 1);

	cout << "grid=(" << grid_dim.x << "," << grid_dim.y << "," << grid_dim.z << ")" << endl;
	cout << "block=(" << block_dim.x << "," << block_dim.y << "," << block_dim.z << ")" << endl;

	c_timer.start();
	g_timer.start();

	// perform computation
	while (parameters.nbr_repeat) {

		// initialize x and y vectors on CPU
		for (int i=0; i<parameters.size; ++i) {
			c_x[i] = static_cast<float>(i+1);
			c_y[i] = static_cast<float>(parameters.size-i);
		}

		// copy x and y resp. to dev_x, dev_y

		cuda_check(cudaMemcpy(g_x, c_x, parameters.size * sizeof(float), cudaMemcpyHostToDevice));
		cuda_check(cudaMemcpy(g_y, c_y, parameters.size * sizeof(float), cudaMemcpyHostToDevice));

		// call kernel
		kernel_sum<<< grid_dim, block_dim >>>(g_x, g_y, g_z, parameters.size);
		cuda_check_kernel();

		// copy result back to computer's global memory
		cuda_check(cudaMemcpy(c_z, g_z, parameters.size * sizeof(float), cudaMemcpyDeviceToHost));

		--parameters.nbr_repeat;
	}

	g_timer.stop();
	c_timer.stop();


	// print vectors if verbose
	if (parameters.verbose) {
		cout << "x = [";
		copy(&c_x[0], &c_x[parameters.size], ostream_iterator<float>(cout, " "));
		cout << "]" << endl << endl;
		cout << "y = [";
		copy(&c_y[0], &c_y[parameters.size], ostream_iterator<float>(cout, " "));
		cout << "]" << endl << endl;
		cout << "z = [";
		copy(&c_z[0], &c_z[parameters.size], ostream_iterator<float>(cout, " "));
		cout << "]" << endl << endl;
	}

	cout << "cpu time(ms)=" << c_timer << endl;
	cout << "gpu time(ms)=" << g_timer << endl;


	// check validity
	float sum = accumulate(&c_z[0], &c_z[parameters.size], 0.0);
	float expected_sum = (parameters.size * (parameters.size+1));

	cout << fixed;
	cout.precision(2);
	cout << "result (found)= " << sum << endl;
	cout << "result (expected)=" << expected_sum << endl;

	assert(sum ==  expected_sum);

	// free memory
	cuda_check(cudaFree( g_x ));
	cuda_check(cudaFree( g_y ));
	cuda_check(cudaFree( g_z ));
	delete [] c_x;
	delete [] c_y;
	delete [] c_z;

	return EXIT_SUCCESS;
}
