/*
 * Copyright 2015 Jean-Michel RICHER
 *
 * This program represents an example of CUDA programming.
 *
 * We compute the sum of two vectors
 *
 * we use UVA Unified Virtual Address space
 */
#include <stdio.h>
#include <cstdlib>
#include <iostream>
#include <algorithm>
#include <numeric>
#include <iterator>
#include <stdexcept>
#include <cassert>
using namespace std;
#include <cuda.h>
#include <getopt.h>

#include "parameters.h"
#include "cpu_timer.h"
#include "gpu_timer.h"

// ==================================================================
// program command line arguments
// ==================================================================
static struct option long_options[] = {
		{"gpu"  	, required_argument, 0, 'g'},
		{"repeat"	, required_argument, 0, 'r'},
		{"size"		, required_argument, 0, 's'},
		{"verbose"	, no_argument, 0, 'v'},
		{0,0,0,0}
};

/**
 * This macro checks return value of the CUDA runtime call and exits
 * the application if the call failed.
 */
#define cuda_check(value) {     \
		cudaError_t err = value; \
		if (err != cudaSuccess) {       \
			cerr << endl; \
			cerr << "============================================\n"; \
			cerr << "Error: " << cudaGetErrorString(err) << " at line "; \
			cerr << __LINE__ << " in file " <<  __FILE__;   \
			cerr <<  endl; \
			exit(EXIT_FAILURE); \
		} \
}

#define cuda_check_kernel() { \
		cudaError_t err = cudaGetLastError(); \
		if (err != cudaSuccess)  { \
			cerr << endl; \
			cerr << "============================================\n"; \
			cerr << "Kernel Error: " << cudaGetErrorString(err) << " at line "; \
			cerr << __LINE__ << " in file " <<  __FILE__;   \
			cerr <<  endl; \
			exit(EXIT_FAILURE); \
		} \
}

__global__ void kernel_sum(float *x, float *y, float *z, int size) {

	int i = threadIdx.x ;
	if (i < size) {
		z[i] = x[i] + y[i]; // paralell part
	}
}


/**
 * main function
 */
int main(int argc, char *argv[]) {
	int gpu_count = -1; // number of devices

	Parameters& parameters = Parameters::get_instance();

	// count the number of devices on the machine
	cuda_check(cudaGetDeviceCount(&gpu_count));

	if (gpu_count == 0) {
		cerr << "No CUDA devices found\n";
		return EXIT_FAILURE;
	}

	parameters.setup(argc, argv, long_options, gpu_count);
	cout << parameters << endl;

	cuda_check(cudaSetDevice(parameters.gpu_id));
	cudaDeviceProp properties;
	cuda_check(cudaGetDeviceProperties(&properties, parameters.gpu_id));


	// allocate data on computer global memory (host) with UVA
	float *u_x, *u_y, *u_z;

	cudaMallocManaged(&u_x, parameters.size * sizeof(float));
	cudaMallocManaged(&u_y, parameters.size * sizeof(float));
	cudaMallocManaged(&u_z, parameters.size * sizeof(float));


	cuda_check(cudaDeviceSynchronize());

	CPUTimer c_timer;
	GPUTimer g_timer;

	dim3 grid_dim(1, 1, 1);
	dim3 block_dim(parameters.size, 1, 1);

	cout << "grid=(" << grid_dim.x << "," << grid_dim.y << "," << grid_dim.z << ")" << endl;
	cout << "block=(" << block_dim.x << "," << block_dim.y << "," << block_dim.z << ")" << endl;

	c_timer.start();
	g_timer.start();

	// perform computation
	while (parameters.nbr_repeat) {

		// initialize x and y vectors
		for (int i=0; i<parameters.size; ++i) {
			u_x[i] = static_cast<float>(i+1);
			u_y[i] = static_cast<float>(parameters.size-i);
		}

		// call kernel
		kernel_sum<<< grid_dim, block_dim >>>(u_x, u_y, u_z, parameters.size);
		// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		// don't forget this with UVA
		cudaDeviceSynchronize();
		// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		cuda_check_kernel();

		--parameters.nbr_repeat;
	}

	g_timer.stop();
	c_timer.stop();


	// print vectors if verbose
	if (parameters.verbose) {
		cout << "x = [";
		copy(&u_x[0], &u_x[parameters.size], ostream_iterator<float>(cout, " "));
		cout << "]" << endl << endl;
		cout << "y = [";
		copy(&u_y[0], &u_y[parameters.size], ostream_iterator<float>(cout, " "));
		cout << "]" << endl << endl;
		cout << "z = [";
		copy(&u_z[0], &u_z[parameters.size], ostream_iterator<float>(cout, " "));
		cout << "]" << endl << endl;
	}

	cout << "cpu time(ms)=" << c_timer << endl;
	cout << "gpu time(ms)=" << g_timer << endl;


	// check validity
	float sum = accumulate(&u_z[0], &u_z[parameters.size], 0.0);
	float expected_sum = (parameters.size * (parameters.size+1));

	cout << fixed;
	cout.precision(2);
	cout << "result (found)= " << sum << endl;
	cout << "result (expected)=" << expected_sum << endl;

	assert(sum ==  expected_sum);

	// free memory
	cuda_check(cudaFree( u_x ));
	cuda_check(cudaFree( u_y ));
	cuda_check(cudaFree( u_z ));

	return EXIT_SUCCESS;
}

